#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Default initialization for dart tests
# Adrien Crovato

from fwk.wutils import parseargs
import dart
import tbox

def mesh(dim, file, pars, bnds, wk = 'wake', wktp = None):
    """Initialize mesh and mesh writer

    Parameters
    ----------
    dim : int (2 or 3)
        number of dimensions
    file : str
        file contaning mesh (w/o extention)
    pars : dict
        parameters for mesh
    bnds : array of str
        names of crack (wake) boundaries physical tag
    wk : str (optional)
        name of crack (wake) physical tag
    wktp : str (optional)
        name of crack (wake) tip physical tag
    """
    import tbox.gmsh as gmsh
    import copy
    # load the mesh
    msh = gmsh.MeshLoader(file,__file__).execute(**pars)
    # crack the mesh
    mshCrck = tbox.MshCrack(msh, dim)
    mshCrck.setCrack(wk)
    for bnd in bnds:
        mshCrck.addBoundary(bnd)
    if dim == 3 and wktp is not None:
        mshCrck.setExcluded(wktp)
    mshCrck.run()
    del mshCrck
    # create writer
    gmshWriter = tbox.GmshExport(msh)
    gmshWriter.save(msh.name)
    return msh, gmshWriter

def problem(msh, dim, alpha, beta, minf, sref, lref, xref, yref, zref, sur, fld = 'field', upstr = 'upstream', ff = 'farfield', dnstr = 'downstream', wk = 'wake', te = 'te', tp = None, vsc = False, dbc = False):
    """Initialize problem

    Parameters
    ----------
    msh : tbox.MshData object
        mesh
    dim : int (2 or 3)
        dimensions
    alpha : float
        angle of attack
    beta : float
        angle of sideslip
    minf : float
        Mach number
    sref : float
        reference area
    lref : float
        reference length
    xref : float
        x-coordinate of reference points for moment computation
    yref : float
        y-coordinate of reference points for moment computation
    zref : float
        z-coordinate of reference points for moment computation
    sur : str
        name of physical tag contaning the body
    fld : str (optional)
        name of physical tag contaning the fluid
    upstr : str (optional)
        name of physical tag contaning the upstream boundary
    ff : str (optional)
        name of physical tag contaning the farfield boundaries
    dnstr : str (optional)
        name of physical tag contaning the downstream boundary
    wk : str (optional)
        name of physical tag contaning the wake
    te : str (optional)
        name of physical tag contaning the trailing edge
    tp : str (optional)
        name of physical tag contaning the wake tip (only for 3D)
    vsc : str (optional)
        name of physical tag contaning the body to apply transpiration B.C.
    dbc : bool (optional)
        True if Dirichlet B.C. are provided explicitely
    """
    # create base formulation
    pbl = dart.Problem(msh, dim, alpha, beta, minf, sref, lref, xref, yref, zref)
    # add incompressible or compressible air fluid
    pbl.set(dart.Fluid(msh, fld, minf, dim, alpha, beta))
    # add initial condition
    pbl.set(dart.Initial(msh, fld, dim, alpha, beta))
    if dbc:
        # Dirichlet BCs provided and explicitely imposed
        dirichlet = dart.Dirichlet(msh, upstr, dim, alpha, beta)
        pbl.add(dirichlet)
    else:
        # Dirichlet BCs not provided, Neumann imposed instead, and a DOF will be pinned automatically
        dirichlet = None
        pbl.add(dart.Freestream(msh, upstr, dim, alpha, beta))
    # add freestream (Neumann) conditions
    pbl.add(dart.Freestream(msh, ff, dim, alpha, beta))
    pbl.add(dart.Freestream(msh, dnstr, dim, alpha, beta))
    # add wake and Kutta conditions
    if tp is None:
        wake = dart.Wake(msh, [wk, wk+'_', fld]) # 2 and 2.5D
    else:
        wake = dart.Wake(msh, [wk, wk+'_', fld, tp]) # 3D
    pbl.add(wake)
    pbl.add(dart.Kutta(msh, [te, wk+'_', sur, fld]))
    # add body of interest
    bnd = dart.Body(msh, [sur, fld])
    pbl.add(bnd)
    # add transpiration BC (VII)
    if vsc:
        blw = dart.Blowing(msh, sur)
        blww = dart.Blowing(msh, wk)
        pbl.add(blw)
        pbl.add(blww)
    else:
        blw = None
        blww = None
    return pbl, dirichlet, wake, bnd, [blw, blww]

def picard(pbl):
    """Initialize Picard solver

    Parameters
    ----------
    pbl : dart.Problem object
        problem formulation
    """
    args = parseargs()
    solver = dart.Picard(pbl, tbox.Gmres(1, 1e-6, 30, 1e-8), nthrds=args.k, vrb=args.verb+2)
    return solver

def newton(pbl):
    """Initialize Newton solver

    Parameters
    ----------
    pbl : dart.Problem object
        problem formulation
    """
    from tbox.solvers import LinearSolver
    args = parseargs()
    solver = dart.Newton(pbl, LinearSolver().pardiso(), nthrds=args.k, vrb=args.verb+2)
    return solver

def morpher(msh, dim, mov, fxd = ['upstream', 'farfield', 'downstream'], fld = 'field', wk = 'wake', sym = 'symmetry'):
    """Initialize mesh morpher

    Parameters
    ----------
    msh : tbox.MshData object
        mesh data structure
    dim : int (2 or 3)
        problem dimensions
    mov : str
        name of physical tag contaning the moving boundary
    fxd : array of str (optional)
        names of physical tags contaning the fixed boundaries
    fld : str (optional)
        name of physical tag contaning the fluid
    wk : str (optional)
        name of physical tag contaning the wake
    sym : str (optional)
        name of physical tag contaning the symmetry plane (y)
    """
    args = parseargs()
    mshDef = tbox.MshDeform(msh, tbox.Gmres(1, 1e-6, 30, 1e-8), dim, nthrds=args.k)
    mshDef.setField(fld)
    for tag in fxd:
        mshDef.addFixed(tag)
    mshDef.addMoving(mov)
    if dim == 3:
        mshDef.setSymmetry(sym, 1)
    mshDef.addInternal(wk, wk+'_')
    mshDef.initialize()
    return mshDef

def adjoint(solver, morpher):
    """Initialize adjoint solver

    Parameters
    ----------
    solver : dart.Newton object
        Newton sovler
    solver : tbox.MshDeform object
        mesh morpher
    """
    adjoint = dart.Adjoint(solver, morpher)
    return adjoint

def initViewer(problem):
    """Initialize viewer

    Parameters
    ----------
    pbl : dart.Problem object
        problem formulation
    """
    args = parseargs()
    if not args.nogui:
        from tbox.viewer import GUI
        GUI().open( problem.msh.name )
